@@ -173,6 +173,36 @@ class UserForm(CaseObject):
         }
         return self.client.request('POST', format_url(self.BASE_URL) + '{}/answerstatus/'.format(pk), json=struct, **kwargs).json()
 
+    def drop_answers(self, pk):
+        answer_dict = self.get_answers(pk)
+        if "answers" in answer_dict and len(answer_dict['answers']) > 0:
+            for key in answer_dict['answers']:
+                answerobj = None
+                if "text" in answer_dict["answers"][key]:
+                    answerobj = ""
+                elif "options" in answer_dict["answers"][key]:
+                    answerobj = []
+                elif "year" in answer_dict["answers"][key]:
+                    answerobj = {
+                        "year": None,
+                        "month": None,
+                        "day": None
+                    }
+                elif "email" in answer_dict["answers"][key]:
+                    answerobj = {
+                        "name": "",
+                        "email": ""
+                    }
+                    
+                ans_obj = {
+                    "fieldtype_id": key.split("-")[1],
+                    "question_id": key.split("-")[0],
+                    "list_id": key.split("-")[2] if key.count("-") == 2 else None,
+                    "client_id": None,
+                    "answer": answerobj
+                }
+                self.save_answer(pk, ans_obj)
+
     def drop_lists(self, pk):
         answer_dict = self.get_answers(pk)
         if "answerlists" in answer_dict and len(answer_dict['answerlists']) > 0:
@@ -189,14 +219,23 @@ class UserForm(CaseObject):
                             "month": None,
                             "day": None
                         }
+                    elif "email" in answer_dict["answers"][key]:
+                        answerobj = {
+                            "name": None,
+                            "email": None
+                        }
                     
-                    self.save_answer(pk, {
-                        "fieldtype_id": key.split("-")[1],
-                        "question_id": key.split("-")[0],
-                        "list_id": key.split("-")[2],
-                        "client_id": None,
-                        "answer": answerobj
-                    })
+                    try:
+                        ans_obj = {
+                            "fieldtype_id": key.split("-")[1],
+                            "question_id": key.split("-")[0],
+                            "list_id": key.split("-")[2],
+                            "client_id": None,
+                            "answer": answerobj
+                        }
+                        self.save_answer(pk, ans_obj)
+                    except:
+                        print("Fail to clear answer?")
             for listobj in answer_dict["answerlists"]:
                 question_id = listobj
                 for listpart in answer_dict["answerlists"][listobj]:
@@ -208,7 +247,7 @@ class UserForm(CaseObject):
         else:
             pass
 
-    def bulk_answer(self, pk, answer_key, options=None):
+    def bulk_answer(self, pk, answer_key, options=None, rewrite_lists=False, conversion_table=None):
         map_list = {}
         for key in answer_key:
             try:
@@ -221,8 +260,20 @@ class UserForm(CaseObject):
                 question_id = clean_key.split("-")[0]
                 fieldtype_id = clean_key.split("-")[1]
                 list_uuid = None
+                list_id = None
                 if "L:" in key:
                     if clean_key.split("-")[2]:
+                        if rewrite_lists:
+                            if conversion_table == None:
+                                raise Exception('No conversion table supplied for list rewrite')
+                            else:
+                                index = int(clean_key.split("-")[2])
+                                new_id = conversion_table['{}-{}'.format(question_id, fieldtype_id)]
+                                nqn = new_id.split('-')[0]
+                                answers = self.get_answers(pk)
+                                answer_list = answers['answerlists']
+                                lid_array = answer_list[nqn]
+                                list_id = lid_array[index - 1]
                         if "key:{}".format(clean_key.split("-")[2]) in map_list:
                             list_uuid = map_list["key:{}".format(clean_key.split("-")[2])]
                         else:
@@ -250,11 +301,18 @@ class UserForm(CaseObject):
                                 answer = option['id']
                     else:
                         raise Exception("Can't lookup answer for option")
+                elif "S:" in key:
+                    p1_name = answer_key[key].split(";")[0]
+                    p1_email = answer_key[key].split(";")[1]
+                    answer = {
+                        'name': p1_name,
+                        'email': p1_email
+                    }
                 elif "C:" in key:
                     if fieldtype_id in options:
                         option_list = options[fieldtype_id]
                         if ";" in answer_key[key]:
-                            answ_arr = answer_key[key].split(';')
+                            answ_arr = answer_key[key].split('; ')
                             answer = []
                             for answerstr in answ_arr:
                                 optFound = False
@@ -278,14 +336,14 @@ class UserForm(CaseObject):
                         raise Exception("Can't lookup answer for option")
                 else:
                     answer = answer_key[key]
-                
-                self.save_answer(pk, {
+                ans_obj = {
                     "fieldtype_clid": fieldtype_id,
                     "question_clid": question_id,
-                    "list_id": None,
-                    "client_id": str(list_uuid) if list_uuid else None,
+                    "list_id": list_id,
+                    "client_id": str(list_uuid) if list_uuid and list_id == None else None,
                     "answer": answer
-                })
+                }
+                self.save_answer(pk, ans_obj)
             except:
                 print("Error: Failed to save an answer; Skipping {} - {}".format(answer_key[key], key))
 
@@ -321,11 +379,40 @@ class UserForm(CaseObject):
     def get_answers(self, pk, **kwargs):
         return self.client.request('GET', format_url(self.BASE_URL) + '{}/answers/'.format(pk), **kwargs).json()
 
-    #downloadEsign
+    def downloadEsign(self, pk, path, eid=None, **kwargs):
+        if eid is None:
+            with self.client.request('GET',
+                            format_url('userforms/userforms/{}/pdf/'.format(
+                                pk
+                            )),
+                            params={
+                                'request': 'esign'
+                            },
+                            stream=True) as response:
+                with open(path, 'wb') as fp:
+                    shutil.copyfileobj(response.raw, fp)
+        else:
+            with self.client.request('GET',
+                            format_url('userforms/userforms/{}/esign/download/{}/'.format(
+                                pk, eid
+                            )),
+                            stream=True) as response:
+                with open(path, 'wb') as fp:
+                    shutil.copyfileobj(response.raw, fp)
 
     #addNote
 
-    #downloadPdf
+    def downloadPdf(self, pk, path, **kwargs):
+        with self.client.request('GET',
+                          format_url('userforms/userforms/{}/pdf/'.format(
+                              pk
+                          )),
+                          params={
+                              'request': 'submission'
+                          },
+                          stream=True) as response:
+            with open(path, 'wb') as fp:
+                shutil.copyfileobj(response.raw, fp)
 
     #downloadCsv
 
@@ -365,6 +452,49 @@ class FieldType(CaseObject):
 class FormTemplate(CaseObject):
     BASE_URL = 'forms/templates/'
 
+    def get_conversion_table(self, pk, to_clid=False):
+        template = self.get(pk)
+
+        conv_table = {}
+
+        def EXPORT_QUESTION(question):
+            question_clid = question['clid']
+            question_id = question['id']
+            
+            if 'atomic_type' in question['fieldtype'] and question['fieldtype']['atomic_type'] is not None:
+                fieldtype_clid = question['fieldtype']['clid']
+                fieldtype_id = question['fieldtype']['id']
+                if to_clid:
+                    conv_table['{}-{}'.format(question_id, fieldtype_id)] = '{}-{}'.format(question_clid, fieldtype_clid)
+                else:
+                    conv_table['{}-{}'.format(question_clid, fieldtype_clid)] = '{}-{}'.format(question_id, fieldtype_id)
+            else:
+                for fieldtype in question['fieldtype']['subfieldtypes']:
+                    fieldtype_clid = fieldtype['clid']
+                    fieldtype_id = fieldtype['id']
+                    if to_clid:
+                        conv_table['{}-{}'.format(question_id, fieldtype_id)] = '{}-{}'.format(question_clid, fieldtype_clid)
+                    else:
+                        conv_table['{}-{}'.format(question_clid, fieldtype_clid)] = '{}-{}'.format(question_id, fieldtype_id)
+
+        def EXPORT_SECTION(parts):
+            for object in parts:
+                if object['archived'] == False:
+                    if object['part_type'] == "q":
+                        EXPORT_QUESTION(object)
+                    elif object['part_type'] == "s":
+                        EXPORT_SECTION(object['parts'])
+
+        def EXPORT_ROOT_CONVERSION(root_parts):
+            for object in root_parts:
+                if object['archived'] == False:
+                    if object['part_type'] == "s":
+                        EXPORT_SECTION(object['parts'])
+
+        EXPORT_ROOT_CONVERSION(template['root_parts'])
+
+        return conv_table
+
     def get_option_array(self, pk):
         id_array = {}
         template = self.get(pk)
@@ -405,6 +535,7 @@ class FormTemplate(CaseObject):
         return id_array
 
     def import_form(self, json):
+        fieldtype_dict = {}
         def CREATE_TEMPLATE(template):
             template_client = FormTemplate(self.client)
             createdTemplate = template_client.post({
@@ -413,143 +544,130 @@ class FormTemplate(CaseObject):
                 'submission_fill_blank': template['submission_prop']['fill_blanks'],
                 'submission_min_rows': template['submission_prop']['hide_blank_rows']
             })
-            print('[DONE] Template `{}` Created!'.format(createdTemplate['name']))
+            for fieldtype in template['fieldtypes']:
+                fieldtype_obj = template['fieldtypes'][fieldtype]
+                if 'subfieldtypes' in fieldtype_obj:
+                    pass
+                else:
+                    fieldtype_id = CREATE_FIELDTYPE(fieldtype_obj)
+                    fieldtype_dict[fieldtype] = fieldtype_id
+            for fieldtype in template['fieldtypes']:
+                fieldtype_obj = template['fieldtypes'][fieldtype]
+                if 'subfieldtypes' in fieldtype_obj:
+                    old_subfields = fieldtype_obj['subfieldtypes']
+                    fieldtype_obj['subfieldtypes'] = []
+                    for subfield in old_subfields:
+                        fieldtype_obj['subfieldtypes'].append(fieldtype_dict[subfield])
+                    fieldtype_id = CREATE_FIELDTYPE(fieldtype_obj)
+                    fieldtype_dict[fieldtype] = fieldtype_id
+                
             CREATE_ROOT_PARTS(template['parts'], createdTemplate['id'])
 
-        def CREATE_ROOT_PARTS(root_parts, templateId):
-            for object in root_parts:
-                if object['type'] == "paragraph":
-                    print('[WAIT] Creating Paragraph~')
-                    CREATE_PARAGRAPH(object, templateId)
-                    print('[DONE] Created Paragraph!')
-                elif object['type'] == "section":
-                    print('[WAIT] Creating Section `{}`~'.format(object['text']))
-                    CREATE_SECTION(object, templateId)
-                    print('[DONE] Created Section `{}`!'.format(object['text']))
-                else:
-                    throw("ERROR: ILLEGAL ROOT PART.")
-
-        def CREATE_PART(parts, templateId, sectionId=None):
-            if parts['type'] == "paragraph":
-                print('[WAIT] Creating Paragraph~')
-                CREATE_PARAGRAPH(parts, templateId, sectionId)
-                print('[DONE] Created Paragraph!')
-            elif parts['type'] == "section":
-                print('[WAIT] Creating Section `{}`~'.format(parts['text']))
-                CREATE_SECTION(parts, templateId, sectionId)
-                print('[DONE] Created Section `{}`!'.format(parts['text']))
-            elif parts['type'] == "question":
-                print('[WAIT] Creating Question `{}`~'.format(parts['text']))
-                CREATE_QUESTION(parts, templateId, sectionId)
-                print('[DONE] Created Question `{}`!'.format(parts['text']))
-            elif parts['type'] == "esign":
-                print('[WAIT] Creating Esign `{}`~'.format(parts['text']))
-                CREATE_ESIGN(parts, templateId, sectionId)
-                print('[DONE] Created Esign `{}`!'.format(parts['text']))
-            else:
-                raise Exception("ERROR: ILLEGAL ROOT PART.")
-
-        def CREATE_QUESTION(question, templateId, sectionId=None):
-            question_client = Question(self.client)
-            print(question['clid'])
+        def CREATE_FIELDTYPE(ftobj):
             fieldtype = {}
-            if 'fieldtype' in question and 'atomic_type' in question['fieldtype']:
-                if question['fieldtype']['atomic_type'] == "text":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('text', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "textarea":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('textarea', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "date":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('date', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "harddate":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('harddate', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "number":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('number', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "currency":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('currency', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "file":
-                    fieldtype = CREATE_BASIC_FIELDTYPE('file', question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "radio":
-                    fieldtype = CREATE_OPTION_FIELDTYPE('radio', question['fieldtype']['options'], question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "checkbox":
-                    fieldtype = CREATE_OPTION_FIELDTYPE('checkbox', question['fieldtype']['options'], question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                elif question['fieldtype']['atomic_type'] == "dropdown":
-                    fieldtype = CREATE_OPTION_FIELDTYPE('dropdown', question['fieldtype']['options'], question['fieldtype']['clid'], question['fieldtype']['required'] if 'required' in question['fieldtype'] else False)
-                question_object = {
-                    'fieldtype': {},
-                    'fieldtype_id': fieldtype['id'],
-                    'is_list': False,
-                    'min_rows': 1,
-                    'number': question['number'] if 'number' in question else None,
-                    'part_type': 'q',
-                    'section': sectionId,
-                    'template': templateId,
-                    'text': question['text'] if 'text' in question else None,
-                    'archived': question['archived'] if 'archived' in question else False,
-                    'clid': question['clid'] if 'clid' in question else None,
-                    'required': question['required'] if 'required' in question else False
-                }
-                question_result = question_client.post(question_object)
-            elif 'fieldtype' in question and 'subfieldtypes' in question['fieldtype']:
-                fieldtype = CREATE_COMPLEX_FIELDTYPE(question['fieldtype']['subfieldtypes'], question['fieldtype']['clid'])
-                question_client.post({
-                    'fieldtype': {},
-                    'fieldtype_id': fieldtype['id'],
-                    'is_list': question['is_list'] if 'is_list' in question else False,
-                    'is_list_table': question['is_list_table'] if 'is_list_table' in question else False,
-                    'min_rows': question['min_rows'] if 'min_rows' in question else 1,
-                    'number': question['number'] if 'number' in question else None,
-                    'part_type': 'q',
-                    'section': sectionId,
-                    'template': templateId,
-                    'text': question['text'] if 'text' in question else None,
-                    'archived': question['archived'] if 'archived' in question else False,
-                    'clid': question['clid'] if 'clid' in question else None,
-                    'required': question['required'] if 'required' in question else False
-                })
+            if 'atomic_type' in ftobj:
+                if ftobj['atomic_type'] in ["text", "textarea", "date", "harddate", "number", "currency", "file", "email", "ssn", "altesign"]:
+                    fieldtype = CREATE_BASIC_FIELDTYPE(ftobj['atomic_type'], ftobj['clid'], ftobj['required'] if 'required' in ftobj else False, ftobj)
+                elif ftobj['atomic_type'] in ["regex"]:
+                    fieldtype = CREATE_REGEX_FIELDTYPE('regex', ftobj['clid'], ftobj['required'] if 'required' in ftobj else False, ftobj)
+                elif ftobj['atomic_type'] in ["radio", "checkbox", "dropdown"]:
+                    fieldtype = CREATE_OPTION_FIELDTYPE(ftobj['atomic_type'], ftobj['options'], ftobj['clid'], ftobj['required'] if 'required' in ftobj else False, ftobj)
+            elif 'subfieldtypes' in ftobj:
+                fieldtype = CREATE_COMPLEX_FIELDTYPE(ftobj['subfieldtypes'], ftobj['clid'])
+            return fieldtype
 
-        def CREATE_BASIC_FIELDTYPE(fieldType, clid=None, required=False):
+        def CREATE_COMPLEX_FIELDTYPE(subFieldTypes, clid=None, required=False):
             fieldtype_client = FieldType(self.client)
-            print('[WAIT] Creating New FieldType `{}`~'.format(fieldType))
-            return fieldtype_client.post({
-                'atomic_type': fieldType,
+            fieldtype_hold = fieldtype_client.post({
+                'atomic_type': None,
                 'option_list': None,
                 'subfieldtypes': [],
                 'clid': clid,
                 'required': required
             })
-
-        def CREATE_COMPLEX_FIELDTYPE(subFieldTypes, clid=None, required=False):
-            fieldtype_client = FieldType(self.client)
-            optionlist_client = OptionList(self.client)
-            print('[WAIT] Creating New Complex FieldType~')
-            subfieldarray = []
+            new_subfieldarray = []
             for subfield in subFieldTypes:
-                appendingArray = {
-                    'atomic_type': subfield['atomic_type'],
-                    'id': None,
-                    'text': subfield['text'] if 'text' in subfield else None,
-                    'clid': subfield['clid'] if 'clid' in subfield else None,
-                    'required': subfield['required'] if 'required' in subfield else False
-                }
-                if 'options' in subfield:
-                    option_list = optionlist_client.post({
-                        'id': None,
-                        'options': subfield['options']
-                    })
-                    appendingArray['option_list'] = option_list['id']
-                subfieldarray.append(appendingArray)
-            return fieldtype_client.post({
+                subfield['parent_type'] = fieldtype_hold['id']
+                fieldtype_client.patch(subfield['id'], subfield)
+                new_subfieldarray.append(subfield)
+            return fieldtype_client.patch(fieldtype_hold['id'], {
+                'id': fieldtype_hold['id'],
                 'atomic_type': None,
                 'option_list': None,
-                'subfieldtypes': subfieldarray,
+                'subfieldtypes': new_subfieldarray,
                 'clid': clid,
                 'required': required
             })
 
-        def CREATE_OPTION_FIELDTYPE(fieldType, optionsToCreate, clid=None, required=False):
+        def CREATE_ROOT_PARTS(root_parts, templateId):
+            for object in root_parts:
+                if object['type'] == "paragraph":
+                    CREATE_PARAGRAPH(object, templateId)
+                elif object['type'] == "section":
+                    CREATE_SECTION(object, templateId)
+                else:
+                    raise Exception("ERROR: Unexpected Root Part")
+
+        def CREATE_PART(parts, templateId, sectionId=None):
+            if parts['type'] == "paragraph":
+                CREATE_PARAGRAPH(parts, templateId, sectionId)
+            elif parts['type'] == "section":
+                CREATE_SECTION(parts, templateId, sectionId)
+            elif parts['type'] == "question":
+                CREATE_QUESTION(parts, templateId, sectionId)
+            elif parts['type'] == "esign":
+                CREATE_ESIGN(parts, templateId, sectionId)
+            else:
+                raise Exception("ERROR: Unexpected Root Part")
+
+        def CREATE_QUESTION(question, templateId, sectionId=None):
+            question_client = Question(self.client)
+            question_client.post({
+                'fieldtype': {},
+                'fieldtype_id': fieldtype_dict[question['fieldtype']]['id'],
+                'is_list': question['is_list'] if 'is_list' in question else False,
+                'is_list_table': question['is_list_table'] if 'is_list_table' in question else False,
+                'min_rows': question['min_rows'] if 'min_rows' in question else 1,
+                'number': question['number'] if 'number' in question else None,
+                'part_type': 'q',
+                'section': sectionId,
+                'template': templateId,
+                'text': question['text'] if 'text' in question else None,
+                'read_only': question['read_only'] if 'read_only' in question else False,
+                'web_only': question['web_only'] if 'web_only' in question else False,
+                'archived': question['archived'] if 'archived' in question else False,
+                'clid': question['clid'] if 'clid' in question else None,
+                'required': question['required'] if 'required' in question else False
+            })
+
+        def CREATE_BASIC_FIELDTYPE(fieldType, clid=None, required=False, ftobj={}):
+            fieldtype_client = FieldType(self.client)
+            return fieldtype_client.post({
+                'atomic_type': fieldType,
+                'option_list': None,
+                'subfieldtypes': [],
+                'clid': clid,
+                'required': required,
+                'text': ftobj['text'] if 'text' in ftobj else None
+            })
+
+        def CREATE_REGEX_FIELDTYPE(fieldType, clid=None, required=False, part={}):
+            fieldtype_client = FieldType(self.client)
+            return fieldtype_client.post({
+                'atomic_type': fieldType,
+                'option_list': None,
+                'subfieldtypes': [],
+                'clid': clid,
+                'required': required,
+                'regex_pattern': part['regex_pattern'],
+                'mask_pattern': part['mask_pattern'],
+                'validation_message': part['validation_message'],
+                'text': part['text'] if 'text' in part else None
+            })
+
+        def CREATE_OPTION_FIELDTYPE(fieldType, optionsToCreate, clid=None, required=False, ftobj={}):
             fieldtype_client = FieldType(self.client)
             optionlist_client = OptionList(self.client)
-            print('[WAIT] Creating New Select FieldType `{}`~'.format(fieldType))
             option_list = optionlist_client.post({
                 'id': None,
                 'options': optionsToCreate
@@ -559,7 +677,8 @@ class FormTemplate(CaseObject):
                 'option_list': option_list['id'],
                 'subfieldtypes': [],
                 'clid': clid,
-                'required': required
+                'required': required,
+                'text': ftobj['text'] if 'text' in ftobj else None
             })
 
         def CREATE_PARAGRAPH(paragraph, templateId, sectionId=None):
@@ -569,6 +688,7 @@ class FormTemplate(CaseObject):
                 'part_type': 'p',
                 'text': paragraph['text'],
                 'archived': paragraph['archived'],
+                'web_only': paragraph['web_only'] if 'web_only' in paragraph else False,
                 'clid': paragraph['clid'] if 'clid' in paragraph else None
             }
             if sectionId is not None:
@@ -589,13 +709,14 @@ class FormTemplate(CaseObject):
 
         def CREATE_SECTION(section, templateId, sectionId=None):
             section_client = Section(self.client)
-            print(section['clid'])
             section_object = {
                 'template': templateId,
                 'part_type': 's',
                 'text': section['text'] if 'text' in section else None,
                 'number': section['number'] if 'number' in section else None,
                 'archived': section['archived'] if 'archived' in section else False,
+                'web_only': section['web_only'] if 'web_only' in section else False,
+                'read_only': section['read_only'] if 'read_only' in section else False,
                 'clid': section['clid'] if 'clid' in section else None
             }
             if sectionId is not None:
