@@ -5,6 +5,6 @@ logging.basicConfig(format="[%(levelname)s]: %(message)s", level=os.environ.get(
 
 def format_url(path):
     if os.environ.get('DEBUG'):
-        return 'http://localhost:{}/v1/'.format(os.environ.get('DEBUG_PORT')) + path
+        return 'http://127.0.0.1:{}/v1/'.format(os.environ.get('DEBUG_PORT')) + path
     else:
         return 'https://{}/api/v1/' + path
